/* bus.c - 6502 microprocessor emulator
 *
 * Copyright (C) 2017, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "xlib/xutil.h"

#include "config.h"
#include "machine.h"
#include "memory.h"
#include "bus.h"


#ifdef DEBUG
#define DEBUG_BUS   0
#endif

#if DEBUG_BUS
extern bool_t mem_quiet;
#endif


void
bus_init (Memory *bus[])
{
	int i;

	for (i = 0; i < ADDRESS_SPACE; i++) {
		bus[i] = NULL;
	}
}


void
bus_attach_device (Memory *bus[], Memory *mem)
{
	int end, i;

	assert (mem->base != -1);

	end = mem->base + mem->size;

	for (i = mem->base; i < end; i++) {
		bus[i] = mem;
	}
}


Byte
bus_read (Memory *bus[], int addr)
{
	Memory *m;

	m = bus[ADDRESS(addr)];

	if (m != NULL) {
		return (mem_get_read_func (m))(m, ADDRESS(addr) - m->base);
	}

#if DEBUG_BUS
	if (!mem_quiet) {
		fprintf (stderr,
		         "%s: bus_read: attempted read from unmapped address %04X\n",
		         PACKAGE, addr);
	}
#endif

#if 1
	/* If out of range, return high byte of address */
	return BYTE_HI (addr);
#else
	return 0;
#endif
}


void
bus_store (Memory *bus[], int addr, Byte b)
{
	Memory *m;

	m = bus[ADDRESS(addr)];

	if (m != NULL) {
		(mem_get_store_func (m))(m, ADDRESS(addr) - m->base, b);
	}
#if DEBUG_BUS
	else if (!mem_quiet) {
		fprintf (stderr,
		         "%s: bus_store: attempted store to unmapped address %04X\n",
		         PACKAGE, addr);
	}
#endif
}


Word
bus_read_word (Memory *bus[], int addr)
{
	Byte hi, lo;

	lo = bus_read (bus, addr);
	hi = bus_read (bus, (addr + 1) & 0xFFFF);

	return WORD_BUILD (hi, lo);
}


Word
bus_read_word_zero (Memory *bus[], int addr)
{
	Byte hi, lo;

	lo = bus_read (bus, addr);
	hi = bus_read (bus, (addr + 1) & 0xFF);

	return WORD_BUILD (hi, lo);
}

