/* cpu.c - 6502 microprocessor emulator
 *
 * Copyright (C) 2017, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"
#include "xlib/xstr.h"

#include "config.h"
#include "opcodes.h"
#include "machine.h"
#include "video.h"
#include "cpu.h"
#include "bus.h"


#ifdef DEBUG
#define DEBUG_CPU       0
#define DEBUG_FLAGS     0
#define DEBUG_STACK     0
#define DEBUG_JSR       0
#endif


#define STACK_BASE  0x0100


#define FLAG_GET(flag)          ((cpu->status & (1 << (flag))) != 0)
#define FLAG_SET(flag, state)   (cpu->status = ((state) ?        \
                                 cpu->status | (1 << (flag)) :   \
                                 cpu->status & ~(1 << (flag))))

#define absolute_addr(cpu)      cpu_fetch_word (cpu)
#define zero_page_addr(cpu)     cpu_fetch_byte (cpu)
#define get_operand(cpu, mode)  bus_read ((cpu)->bus, get_address (cpu, mode))


typedef enum {

	AM_ACCUMULATOR,
	AM_IMMEDIATE,
	AM_RELATIVE,
	AM_ABSOLUTE,
	AM_ABSOLUTE_X,
	AM_ABSOLUTE_Y,
	AM_ZERO_PAGE,
	AM_ZERO_PAGE_X,
	AM_ZERO_PAGE_Y,
	AM_INDIRECT,
	AM_INDIRECT_X,
	AM_INDIRECT_Y

} AddressMode;


typedef enum {

	REG_A,
	REG_X,
	REG_Y

} Register;


typedef enum {
	      /* bit # */
	CPU_FLAG_C = 0,    /* 0x01  carry             */
	CPU_FLAG_Z = 1,    /* 0x02  zero              */
	CPU_FLAG_I = 2,    /* 0x04  interrupt disable */
	CPU_FLAG_D = 3,    /* 0x08  BCD               */
	CPU_FLAG_B = 4,    /* 0x10  breakpoint        */
	CPU_FLAG_x = 5,    /* 0x20  unused            */
	CPU_FLAG_V = 6,    /* 0x40  overflow          */
	CPU_FLAG_N = 7     /* 0x80  negative          */

} CpuStatus;


extern Machine *machine;


struct _CPU {

	const char *progname;

	CpuType type;

	Byte a;
	Byte x;
	Byte y;

	Word pc;

	Byte sp; /* low byte of stack pointer; high byte is hardwired to 0x01 */

	Byte status;

	xbool_t signal_reset;
	xbool_t signal_nmi;
	xbool_t signal_irq;

	xbool_t running;

	xbool_t signal_toggle_stepping;
	xbool_t signal_step_now;
	xbool_t stepping;

	int vector_reset;
	int vector_nmi;
	int vector_irq;

	Memory **bus;

};


static int bcd[] = {
	  0,   1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12,  13,  14,  15,
	 10,  11,  12,  13,  14,  15,  16,  17,  18,  19,  20,  21,  22,  23,  24,  25,
	 20,  21,  22,  23,  24,  25,  26,  27,  28,  29,  30,  31,  32,  33,  34,  35,
	 30,  31,  32,  33,  34,  35,  36,  37,  38,  39,  40,  41,  42,  43,  44,  45,
	 40,  41,  42,  43,  44,  45,  46,  47,  48,  49,  50,  51,  52,  53,  54,  55,
	 50,  51,  52,  53,  54,  55,  56,  57,  58,  59,  60,  61,  62,  63,  64,  65,
	 60,  61,  62,  63,  64,  65,  66,  67,  68,  69,  70,  71,  72,  73,  74,  75,
	 70,  71,  72,  73,  74,  75,  76,  77,  78,  79,  80,  81,  82,  83,  84,  85,
	 80,  81,  82,  83,  84,  85,  86,  87,  88,  89,  90,  91,  92,  93,  94,  95,
	 90,  91,  92,  93,  94,  95,  96,  97,  98,  99, 100, 101, 102, 103, 104, 105,
	100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115,
	110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125,
	120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132, 133, 134, 135,
	130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145,
	140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155,
	150, 151, 152, 153, 154, 155, 156, 157, 158, 159, 160, 161, 162, 163, 164, 165
};


static xbool_t cpu_execute (CPU *cpu, Byte op);


static void
cpu_fatal (CPU *cpu, char *message)
{
	if (!message) {
		message = xstr_dup ("Unspecified error - can't continue");
	}
	fprintf (stderr, "%s: CPU: %s\n", cpu->progname, message);
	xfree (message);
	exit (EXIT_FAILURE);
}


static void
cpu_reset (CPU *cpu)
{
	cpu->a = 0x00;
	cpu->x = 0x00;
	cpu->y = 0x00;

	cpu->status = 0x04 | 0x20;

	cpu->sp = 0xFF;

	cpu->pc = 0x0000;

	cpu->signal_reset = FALSE;
	cpu->signal_nmi   = FALSE;
	cpu->signal_irq   = FALSE;

	cpu->running = TRUE;
}


CPU*
cpu_new (Machine *m, const char *progname, Memory *bus[], CpuType type,
	int vector_reset, int vector_nmi, int vector_irq)
{
	CPU *cpu;

	machine = m;

	cpu = xnew0 (CPU, 1);

	cpu->progname = progname;
	cpu->bus  = bus;
	cpu->type = type;

	cpu->vector_reset = vector_reset;
	cpu->vector_nmi   = vector_nmi;
	cpu->vector_irq   = vector_irq;

	cpu_reset (cpu);

	return cpu;
}


void
cpu_free (CPU *cpu)
{
	xfree (cpu);
}


void
cpu_status_dump (FILE *f, CPU *cpu)
{
	fprintf (f, "                ");

	fprintf (f, "PC = %04X", cpu->pc);
	fprintf (f, ", A = %02X", cpu->a);
	fprintf (f, ", X = %02X", cpu->x);
	fprintf (f, ", Y = %02X", cpu->y);
	fprintf (f, ", SP = 01%02X", cpu->sp);
	fprintf (f, ", PS = ");
	fprintf (f, "%c", FLAG_GET (CPU_FLAG_N) ? 'N' : 'n');
	fprintf (f, "%c", FLAG_GET (CPU_FLAG_V) ? 'V' : 'v');
	fprintf (f, "%c", FLAG_GET (CPU_FLAG_x) ? '+' : '-');
	fprintf (f, "%c", FLAG_GET (CPU_FLAG_B) ? 'B' : 'b');
	fprintf (f, "%c", FLAG_GET (CPU_FLAG_D) ? 'D' : 'd');
	fprintf (f, "%c", FLAG_GET (CPU_FLAG_I) ? 'I' : 'i');
	fprintf (f, "%c", FLAG_GET (CPU_FLAG_Z) ? 'Z' : 'z');
	fprintf (f, "%c", FLAG_GET (CPU_FLAG_C) ? 'C' : 'c');

	fprintf (f, "\n");
}


static Byte
cpu_reg_read (CPU *cpu, Register reg)
{
	switch (reg) {
		case REG_A: return cpu->a;
		case REG_X: return cpu->x;
		case REG_Y: return cpu->y;
		default: assert (NULL); return 0;
	}
}


static void
cpu_set_N_Z (CPU *cpu, Byte b)
{
	FLAG_SET (CPU_FLAG_Z, b == 0);
	FLAG_SET (CPU_FLAG_N, (b & 0x80) != 0);
}


static void
cpu_reg_set_with_N_Z (CPU *cpu, Register reg, Byte b)
{
	switch (reg) {
		case REG_A: cpu->a = b; break;
		case REG_X: cpu->x = b; break;
		case REG_Y: cpu->y = b; break;
	}

	FLAG_SET (CPU_FLAG_Z, b == 0);
	FLAG_SET (CPU_FLAG_N, (b & 0x80) != 0);
}


static void
cpu_compare (CPU *cpu, Byte a, Byte b)
{
	FLAG_SET (CPU_FLAG_N, a < b);
	FLAG_SET (CPU_FLAG_Z, a == b);
	FLAG_SET (CPU_FLAG_C, a >= b);
}


static Byte
cpu_inc (CPU *cpu, Byte b)
{
	b++;

	FLAG_SET (CPU_FLAG_Z, b == 0);
	FLAG_SET (CPU_FLAG_N, (b & 0x80) != 0);

	return b;
}


static Byte
cpu_dec (CPU *cpu, Byte b)
{
	b--;

	FLAG_SET (CPU_FLAG_Z, b == 0);
	FLAG_SET (CPU_FLAG_N, (b & 0x80) != 0);

	return b;
}


static Byte
cpu_adc (CPU *cpu, Byte a, Byte b)
{
	unsigned int resulti;
	Byte resultb;

	if (FLAG_GET (CPU_FLAG_D)) {
		/* Decimal mode. */
#if 0
		cpu_fatal (cpu, xstr_dup ("BCD arithmetic not yet implemented"));
#else
		resulti = bcd[a] + bcd[b] + (FLAG_GET (CPU_FLAG_C) ? 1 : 0);
		resultb = (Byte)resulti;

		/* XXX: Check and test against the examples here:
		   See: http://www.6502.org/tutorials/decimal_mode.html */

		FLAG_SET (CPU_FLAG_Z, resultb == 0);
		FLAG_SET (CPU_FLAG_N, (resultb & 0x80) != 0);
		FLAG_SET (CPU_FLAG_C, resulti < 0x00 || resulti > 99);
		FLAG_SET (CPU_FLAG_V, ((resultb & 0x80) != (a & 0x80)));
#endif
	}
	else {
		/* Binary mode. */
		resulti = a + b + (FLAG_GET (CPU_FLAG_C) ? 1 : 0);
		resultb = (Byte)resulti;

		FLAG_SET (CPU_FLAG_Z, resultb == 0);
		FLAG_SET (CPU_FLAG_N, (resultb & 0x80) != 0);
		FLAG_SET (CPU_FLAG_C, resulti < 0x00 || resulti > 0xFF);
		/* V if sign change */
		FLAG_SET (CPU_FLAG_V, ((resultb & 0x80) != (a & 0x80)));
	}

	return resultb;
}


static Byte
cpu_sbc (CPU *cpu, Byte a, Byte b)
{
	unsigned int resulti;
	Byte resultb;

	if (FLAG_GET (CPU_FLAG_D)) {
		/* Decimal mode. */
#if 0
		cpu_fatal (cpu, xstr_dup ("BCD arithmetic not yet implemented"));
#else
		resulti = bcd[a] - bcd[b] - (FLAG_GET (CPU_FLAG_C) ? 0 : 1);
		resultb = (Byte)resulti;

		/* XXX: Check and test against the examples here:
		   See: http://www.6502.org/tutorials/decimal_mode.html */

		FLAG_SET (CPU_FLAG_Z, resultb == 0);
		FLAG_SET (CPU_FLAG_N, (resultb & 0x80) != 0);
		FLAG_SET (CPU_FLAG_C, resulti >= 0x00 && resulti <= 0xFF);
		/* V if sign change */
		FLAG_SET (CPU_FLAG_V, ((resultb & 0x80) != (a & 0x80)));
#endif
	}
	else {
		/* Binary mode. */
		resulti = a - b - (FLAG_GET (CPU_FLAG_C) ? 0 : 1);
		resultb = (Byte)resulti;

		FLAG_SET (CPU_FLAG_Z, resultb == 0);
		FLAG_SET (CPU_FLAG_N, (resultb & 0x80) != 0);
		FLAG_SET (CPU_FLAG_C, resulti >= 0x00 && resulti <= 0xFF);
		/* V if sign change */
		FLAG_SET (CPU_FLAG_V, ((resultb & 0x80) != (a & 0x80)));
	}

	return resultb;
}


static Byte
cpu_rol (CPU *cpu, Byte val, xbool_t carry)
{
	FLAG_SET (CPU_FLAG_C, (val & 0x80) != 0);
	return (Byte)((val << 1) | (carry ? 1 : 0));
}


static Byte
cpu_ror (CPU *cpu, Byte val, xbool_t carry)
{
	FLAG_SET (CPU_FLAG_C, (val & 1) != 0);
	return (Byte)(((val & 0xFF) >> 1) | (carry ? 0x80 : 0));
}


static Byte
cpu_shift_left (CPU *cpu, Byte val)
{
	FLAG_SET (CPU_FLAG_C, (val & 0x80) != 0);
	return (Byte)(val << 1);
}


static Byte
cpu_shift_right (CPU *cpu, Byte val)
{
	FLAG_SET (CPU_FLAG_C, (val & 1) != 0);
	return (Byte)((val & 0xFF) >> 1);
}


static Byte
cpu_fetch_byte (CPU *cpu)
{
	Byte b;

	b = bus_read (cpu->bus, cpu->pc);
	cpu->pc++;

#if DEBUG_CPU
	fprintf (stderr, "%02X ", b);
#endif

	return b;
}


static Word
cpu_fetch_word (CPU *cpu)
{
	Word w;

	w = bus_read_word (cpu->bus, cpu->pc);
	cpu->pc = cpu->pc + 2;

#if DEBUG_CPU
	fprintf (stderr, "%02X %02X", BYTE_LO(w), BYTE_HI(w));
#endif

	return w;
}


static void
cpu_stack_push (CPU *cpu, Byte b)
{
#if DEBUG_STACK
	fprintf (stderr, "\tpushing %02X to %04X\n", b,  STACK_BASE + cpu->sp);
#endif

	bus_store (cpu->bus, STACK_BASE + cpu->sp, b);

	if (cpu->sp == 0x00) {
		fprintf (stderr, "%s: 6502 CPU stack overflow\n", PACKAGE);
	}

	cpu->sp--;
}


static Byte
cpu_stack_pull (CPU *cpu)
{
	Byte b;

	cpu->sp++;

	if (cpu->sp == 0x00) {
		fprintf (stderr, "%s: 6502 CPU stack underflow\n", PACKAGE);
	}

	b = bus_read (cpu->bus, STACK_BASE + cpu->sp);

#if DEBUG_STACK
	fprintf (stderr, "\tpulled %02X from %04X\n", b, STACK_BASE + cpu->sp);
#endif

	return b;
}


static void
cpu_stack_push_word (CPU *cpu, Word w)
{
	cpu_stack_push (cpu, BYTE_HI (w));
	cpu_stack_push (cpu, BYTE_LO (w));
}


static Word
cpu_stack_pull_word (CPU *cpu)
{
	Byte hi, lo;

	lo = cpu_stack_pull (cpu);
	hi = cpu_stack_pull (cpu);

	return WORD_BUILD (hi, lo);
}


static Word
absolute_addr_X (CPU *cpu, Byte offset)
{
	return absolute_addr (cpu) + offset;
}


static Word
zero_page_addr_X (CPU *cpu, Byte offset)
{
	return (zero_page_addr (cpu) + offset) & 0xFF;
}


static Word
indirect_addr (CPU *cpu)
{
	return bus_read_word (cpu->bus, absolute_addr (cpu));
}


static Word
indirect_y_addr (CPU *cpu)
{
	return bus_read_word_zero (cpu->bus, zero_page_addr (cpu)) + cpu->y;
}


static Word
indirect_x_addr (CPU *cpu)
{
	return bus_read_word_zero (cpu->bus,
	                           (zero_page_addr (cpu) + cpu->x) & 0xFF);
}


static Word
relative_addr (CPU *cpu)
{
	return (signed char)cpu_fetch_byte (cpu) + cpu->pc;
}


static Word
get_address (CPU *cpu, AddressMode mode)
{
	switch (mode) {
	case AM_ABSOLUTE:
		return ADDRESS (absolute_addr (cpu));
	case AM_RELATIVE:
		return ADDRESS (relative_addr (cpu));
	case AM_ABSOLUTE_X:
		return ADDRESS (absolute_addr_X (cpu, cpu->x));
	case AM_ABSOLUTE_Y:
		return ADDRESS (absolute_addr_X (cpu, cpu->y));
	case AM_ZERO_PAGE:
		return ADDRESS (zero_page_addr (cpu));
	case AM_ZERO_PAGE_X:
		return ADDRESS (zero_page_addr_X (cpu, cpu->x));
	case AM_ZERO_PAGE_Y:
		return ADDRESS (zero_page_addr_X (cpu, cpu->y));
	case AM_INDIRECT_X:
		return ADDRESS (indirect_x_addr (cpu));
	case AM_INDIRECT_Y:
		return ADDRESS (indirect_y_addr (cpu));
	case AM_INDIRECT:
		return ADDRESS (indirect_addr (cpu));

	case AM_IMMEDIATE:
	case AM_ACCUMULATOR:
	default:
		assert (NULL);
		break;
	}
}


static void
op_adc (CPU *cpu, AddressMode mode)
{
	if (mode == AM_IMMEDIATE) {
		cpu->a = cpu_adc (cpu, cpu->a, cpu_fetch_byte (cpu));
	}
	else {
		cpu->a = cpu_adc (cpu, cpu->a, get_operand (cpu, mode));
	}
}


static void
op_and (CPU *cpu, AddressMode mode)
{
	if (mode == AM_IMMEDIATE) {
		cpu_reg_set_with_N_Z (cpu, REG_A, cpu->a & cpu_fetch_byte (cpu));
	}
	else {
		cpu_reg_set_with_N_Z (cpu, REG_A, cpu->a & get_operand (cpu, mode));
	}
}


static void
op_asl (CPU *cpu, AddressMode mode)
{
	if (mode == AM_ACCUMULATOR) {
		cpu_reg_set_with_N_Z (cpu, REG_A, cpu_shift_left (cpu, cpu->a));
	}
	else {
		Word addr = get_address (cpu, mode);
		unsigned int result = cpu_shift_left (cpu, bus_read (cpu->bus, addr));
		bus_store (cpu->bus, addr, result);
		cpu_set_N_Z (cpu, result);		
	}
}


static void
op_bit (CPU *cpu, AddressMode mode)
{
	Byte b;

	b = get_operand (cpu, mode);

	FLAG_SET (CPU_FLAG_Z, (cpu->a & b) == 0);
	FLAG_SET (CPU_FLAG_V, (b & 0x40) != 0);
	FLAG_SET (CPU_FLAG_N, (b & 0x80) != 0);
}


static void
op_bXs (CPU *cpu, CpuStatus flag, xbool_t test)
{
	Word addr;

	addr = get_address (cpu, AM_RELATIVE);
	if (FLAG_GET (flag) == test) {
		cpu->pc = addr;
	}
}


static void
op_brk (CPU *cpu)
{
	/* XXX */
	fprintf (stderr, "*** BRK: Hit <ENTER> to continue...\n");
	cpu_status_dump (stderr, cpu);
	fgetc (stdin);

#if 0
	FLAG_SET (CPU_FLAG_B, TRUE);
	FLAG_SET (CPU_FLAG_I, TRUE);

	cpu_stack_push_word (cpu, cpu->pc + 1);
	cpu_stack_push (cpu, cpu->status | 0x10);

	cpu->pc = bus_read (cpu->bus, (Word)cpu->vector_irq);
#endif
}


static void
op_cmp (CPU *cpu, AddressMode mode)
{
	if (mode == AM_IMMEDIATE) {
		cpu_compare (cpu, cpu->a, cpu_fetch_byte (cpu));
	}
	else {
		cpu_compare (cpu, cpu->a, get_operand (cpu, mode));
	}
}


static void
op_cpX (CPU *cpu, Byte b, AddressMode mode)
{
	if (mode == AM_IMMEDIATE) {
		cpu_compare (cpu, b, cpu_fetch_byte (cpu));
	}
	else {
		cpu_compare (cpu, b, get_operand (cpu, mode));
	}
}


static void
op_dec (CPU *cpu, AddressMode mode)
{
	Word addr;

	addr = get_address (cpu, mode);
	bus_store (cpu->bus, addr, cpu_dec (cpu, bus_read (cpu->bus, addr)));
}


static void
op_deX (CPU *cpu, Register reg)
{
	switch (reg) {
		case REG_X: cpu->x = cpu_dec (cpu, cpu->x); break;
		case REG_Y: cpu->y = cpu_dec (cpu, cpu->y); break;
		default: assert (NULL); break;
	}
}


static void
op_eor (CPU *cpu, AddressMode mode)
{
	if (mode == AM_IMMEDIATE) {
		cpu_reg_set_with_N_Z (cpu, REG_A, cpu->a ^ cpu_fetch_byte (cpu));
	}
	else {
		cpu_reg_set_with_N_Z (cpu, REG_A, cpu->a ^ get_operand (cpu, mode));
	}
}


static void
op_inc (CPU *cpu, AddressMode mode)
{
	Word addr;

	addr = get_address (cpu, mode);
	bus_store (cpu->bus, addr, cpu_inc (cpu, bus_read (cpu->bus, addr)));
}


static void
op_inX (CPU *cpu, Register reg)
{
	switch (reg) {
		case REG_X: cpu->x = cpu_inc (cpu, cpu->x); break;
		case REG_Y: cpu->y = cpu_inc (cpu, cpu->y); break;
		default: assert (NULL); break;
	}
}


static void
op_jmp (CPU *cpu, AddressMode mode)
{
	cpu->pc = get_address (cpu, mode);
}


#if 0
static void
op_jmp_indirect (CPU *cpu)
{
	cpu->pc = bus_read_word (cpu->bus, get_address (cpu, AM_ABSOLUTE));
}
#endif


static void
op_jsr (CPU *cpu)
{
#if DEBUG_JSR
	fprintf (stderr, "JSR %02X %02X\n",
	         BYTE_LO (cpu->pc + 1), BYTE_HI (cpu->pc + 1));
#endif

	cpu_stack_push_word (cpu, cpu->pc + 1);
	cpu->pc = get_address (cpu, AM_ABSOLUTE);
}


static void
op_ldX (CPU *cpu, Register reg, AddressMode mode)
{
	if (mode == AM_IMMEDIATE) {
		cpu_reg_set_with_N_Z (cpu, reg, cpu_fetch_byte (cpu));
	}
	else {
		cpu_reg_set_with_N_Z (cpu, reg, get_operand (cpu, mode));
	}
}


static void
op_lsr (CPU *cpu, AddressMode mode)
{
	if (mode == AM_ACCUMULATOR) {
		cpu_reg_set_with_N_Z (cpu, REG_A, cpu_shift_right (cpu, cpu->a));
	}
	else {
		Word addr = get_address (cpu, mode);
		unsigned int result = cpu_shift_right (cpu, bus_read (cpu->bus, addr));
		bus_store (cpu->bus, addr, result);
		cpu_set_N_Z (cpu, result);
	}
}


static xbool_t
op_nop (void)
{
	return TRUE;
}


static void
op_ora (CPU *cpu, AddressMode mode)
{
	if (mode == AM_IMMEDIATE) {
		cpu_reg_set_with_N_Z (cpu, REG_A, cpu->a | cpu_fetch_byte (cpu));
	}
	else {
		cpu_reg_set_with_N_Z (cpu, REG_A, cpu->a | get_operand (cpu, mode));
	}
}


static void
op_pha (CPU *cpu)
{
	cpu_stack_push (cpu, cpu->a);
}


static void
op_pla (CPU *cpu)
{
	cpu_reg_set_with_N_Z (cpu, REG_A, cpu_stack_pull (cpu));
}


static void
op_php (CPU *cpu)
{
#if 0
	fprintf (stderr, "cpu status = [%d]\n", cpu->status);
#endif
	cpu_stack_push (cpu, cpu->status);
}


static void
op_plp (CPU *cpu)
{
	cpu->status = cpu_stack_pull (cpu) | 0x20 | 0x10; /* XXX: checkme */
#if 0
	fprintf (stderr, "cpu status = [%d]\n", cpu->status);
#endif
}


static void
op_rol (CPU *cpu, int mode)
{
	if (mode == AM_ACCUMULATOR) {
		cpu_reg_set_with_N_Z (cpu, REG_A,
		    cpu_rol (cpu, cpu->a, FLAG_GET (CPU_FLAG_C)));
	}
	else {
		Word addr = get_address (cpu, mode);
		unsigned int result = cpu_rol (cpu,
		    bus_read (cpu->bus, addr), FLAG_GET (CPU_FLAG_C));
		bus_store (cpu->bus, addr, result);
		cpu_set_N_Z (cpu, result);
	}
}


static void
op_ror (CPU *cpu, int mode)
{
	if (mode == AM_ACCUMULATOR) {
		cpu_reg_set_with_N_Z (cpu, REG_A,
		    cpu_ror (cpu, cpu->a, FLAG_GET (CPU_FLAG_C)));
	}
	else {
		Word addr = get_address (cpu, mode);
		unsigned int result = cpu_ror (cpu,
		    bus_read (cpu->bus, addr), FLAG_GET (CPU_FLAG_C));
		bus_store (cpu->bus, addr, result);
		cpu_set_N_Z (cpu, result);
	}
}


static xbool_t
op_rts (CPU *cpu)
{
	if (cpu->sp == 0xFF) {
		return FALSE;
	}

	cpu->pc = ADDRESS (cpu_stack_pull_word (cpu) + 1);

#if DEBUG_JSR
	fprintf (stderr, "RTS [%04X]\n", cpu->pc);
#endif

	return TRUE;
}


static void
op_rti (CPU *cpu)
{
	cpu->status = cpu_stack_pull (cpu);
	cpu->pc = cpu_stack_pull_word (cpu);
}


static void
op_sbc (CPU *cpu, AddressMode mode)
{
	if (mode == AM_IMMEDIATE) {
		cpu->a = cpu_sbc (cpu, cpu->a, cpu_fetch_byte (cpu));
	}
	else {
		cpu->a = cpu_sbc (cpu, cpu->a, get_operand (cpu, mode));
	}
}


static void
op_stX (CPU *cpu, Register reg, AddressMode mode)
{
	bus_store (cpu->bus, get_address (cpu, mode),
	    cpu_reg_read (cpu, reg));
}


static void
op_taX (CPU *cpu, Register reg)
{
	cpu_reg_set_with_N_Z (cpu, reg, cpu->a);
}


static void
op_tsx (CPU *cpu)
{
	cpu_reg_set_with_N_Z (cpu, REG_X, cpu->sp);
}


static void
op_txs (CPU *cpu)
{
	cpu->sp = cpu->x;
#if 0
	cpu_set_N_Z (cpu, cpu->sp);  /* XXX: checkme - should txs really set these flags? */
#endif
}


static void
op_tXa (CPU *cpu, Register reg)
{
	cpu_reg_set_with_N_Z (cpu, REG_A, cpu_reg_read (cpu, reg));
}


static void
op_not_implemented (CPU *cpu, Byte op)
{
	cpu_status_dump (stderr, cpu);
	cpu_fatal (cpu, xstr_dup_printf ("Illegal/unimplemented operation %02X", op));
}


static void
signal_handler (CPU *cpu)
{
	/* 6502 signals */

	if (cpu->signal_reset) {

		if (cpu->vector_reset == -1) {
			cpu->signal_reset = FALSE;
		}

		cpu_reset (cpu);
		cpu->pc = bus_read_word (cpu->bus, cpu->vector_reset);

	}

	if (cpu->signal_nmi) {

		cpu->signal_nmi = FALSE;

		if (cpu->vector_nmi != -1) {
			cpu_stack_push_word (cpu, cpu->pc);
			cpu_stack_push (cpu, cpu->status);
			FLAG_SET (CPU_FLAG_I, TRUE);
			cpu->pc = bus_read_word (cpu->bus, cpu->vector_nmi);
		}

	}

	if (cpu->signal_irq) {

		cpu->signal_irq = FALSE;

		if (cpu->vector_irq != -1) {
			cpu_stack_push_word (cpu, cpu->pc);
			cpu_stack_push (cpu, cpu->status & ~0x10);
			FLAG_SET (CPU_FLAG_I, TRUE);
			cpu->pc = bus_read_word (cpu->bus, cpu->vector_irq);
		}

	}

	/* Emulator signals */

	if (cpu->signal_toggle_stepping) {

		cpu->signal_toggle_stepping = FALSE;
		cpu->stepping = !cpu->stepping;
		cpu->running = ! cpu->stepping;

	}

	if (!cpu->running && cpu->signal_step_now) {

		cpu->signal_step_now = FALSE;
		cpu->running = TRUE;

	}


}


void
cpu_signal_reset (CPU *cpu)
{
	cpu->signal_reset = TRUE;
}


void
cpu_signal_toggle_stepping (CPU *cpu)
{
	cpu->signal_toggle_stepping = TRUE;
}


void
cpu_signal_step_now (CPU *cpu)
{
	cpu->signal_step_now = TRUE;
}


xbool_t
cpu_run (CPU *cpu, int origin)
{
	/* struct timespec sleep_time; */
	Byte op;
	xbool_t quit = FALSE;

	/*
	sleep_time.tv_sec = 0;
	sleep_time.tv_nsec = 2;
	*/

	cpu->pc = ADDRESS (origin);

	while (!quit) {

			video_refresh ();

			signal_handler (cpu);
			environment_event_handler (&quit);

			if (cpu->running) {

#if DEBUG_CPU
				fprintf (stderr, "%04X: ", cpu->pc);
#endif

				op = cpu_fetch_byte (cpu);

				if (!cpu_execute (cpu, op)) {
					break;
				}


#if DEBUG_CPU
				fprintf (stderr, "\n");
#endif

#if DEBUG_FLAGS
				cpu_status_dump (stderr, cpu);
#endif

			}

			if (cpu->stepping) {
				cpu->running = FALSE;
			}

			/* nanosleep (&sleep_time, NULL); */
	}

	return TRUE;
}


static xbool_t
cpu_execute (CPU *cpu, Byte op)
{

	switch (op) {

		case OP_ADC:     op_adc (cpu, AM_IMMEDIATE); break;
		case OP_ADC_A:   op_adc (cpu, AM_ABSOLUTE); break;
		case OP_ADC_AX:  op_adc (cpu, AM_ABSOLUTE_X); break;
		case OP_ADC_AY:  op_adc (cpu, AM_ABSOLUTE_Y); break;
		case OP_ADC_Z:   op_adc (cpu, AM_ZERO_PAGE); break;
		case OP_ADC_ZX:  op_adc (cpu, AM_ZERO_PAGE_X); break;
		case OP_ADC_IX:  op_adc (cpu, AM_INDIRECT_X); break;
		case OP_ADC_IY:  op_adc (cpu, AM_INDIRECT_Y); break;

		case OP_AND:     op_and (cpu, AM_IMMEDIATE); break;
		case OP_AND_A:   op_and (cpu, AM_ABSOLUTE); break;
		case OP_AND_AX:  op_and (cpu, AM_ABSOLUTE_X); break;
		case OP_AND_AY:  op_and (cpu, AM_ABSOLUTE_Y); break;
		case OP_AND_Z:   op_and (cpu, AM_ZERO_PAGE); break;
		case OP_AND_ZX:  op_and (cpu, AM_ZERO_PAGE_X); break;
		case OP_AND_IX:  op_and (cpu, AM_INDIRECT_X); break;
		case OP_AND_IY:  op_and (cpu, AM_INDIRECT_Y); break;

		case OP_ASL_ACC: op_asl (cpu, AM_ACCUMULATOR); break;
		case OP_ASL_A:   op_asl (cpu, AM_ABSOLUTE); break;
		case OP_ASL_AX:  op_asl (cpu, AM_ABSOLUTE_X); break;
		case OP_ASL_Z:   op_asl (cpu, AM_ZERO_PAGE); break;
		case OP_ASL_ZX:  op_asl (cpu, AM_ZERO_PAGE_X); break;

		case OP_BCS_R:   op_bXs (cpu, CPU_FLAG_C, TRUE); break;
		case OP_BCC_R:   op_bXs (cpu, CPU_FLAG_C, FALSE); break;
		case OP_BEQ_R:   op_bXs (cpu, CPU_FLAG_Z, TRUE); break;
		case OP_BNE_R:   op_bXs (cpu, CPU_FLAG_Z, FALSE); break;
		case OP_BMI_R:   op_bXs (cpu, CPU_FLAG_N, TRUE); break;
		case OP_BPL_R:   op_bXs (cpu, CPU_FLAG_N, FALSE); break;
		case OP_BVS_R:   op_bXs (cpu, CPU_FLAG_V, TRUE); break;
		case OP_BVC_R:   op_bXs (cpu, CPU_FLAG_V, FALSE); break;

		case OP_BIT_A:   op_bit (cpu, AM_ABSOLUTE); break;
		case OP_BIT_Z:   op_bit (cpu, AM_ZERO_PAGE); break;

		case OP_BRK_IMP: op_brk (cpu); break;

		case OP_CLC_IMP: FLAG_SET (CPU_FLAG_C, FALSE); break;
		case OP_CLV_IMP: FLAG_SET (CPU_FLAG_V, FALSE); break;
		case OP_CLI_IMP: FLAG_SET (CPU_FLAG_I, FALSE); break;
		case OP_CLD_IMP: FLAG_SET (CPU_FLAG_D, FALSE); break;

		case OP_CMP:     op_cmp (cpu, AM_IMMEDIATE); break;
		case OP_CMP_A:   op_cmp (cpu, AM_ABSOLUTE); break;
		case OP_CMP_AX:  op_cmp (cpu, AM_ABSOLUTE_X); break;
		case OP_CMP_AY:  op_cmp (cpu, AM_ABSOLUTE_Y); break;
		case OP_CMP_Z:   op_cmp (cpu, AM_ZERO_PAGE); break;
		case OP_CMP_ZX:  op_cmp (cpu, AM_ZERO_PAGE_X); break;
		case OP_CMP_IX:  op_cmp (cpu, AM_INDIRECT_X); break;
		case OP_CMP_IY:  op_cmp (cpu, AM_INDIRECT_Y); break;
		case OP_CPX:     op_cpX (cpu, cpu->x, AM_IMMEDIATE); break;
		case OP_CPX_A:   op_cpX (cpu, cpu->x, AM_ABSOLUTE); break;
		case OP_CPX_Z:   op_cpX (cpu, cpu->x, AM_ZERO_PAGE); break;
		case OP_CPY:     op_cpX (cpu, cpu->y, AM_IMMEDIATE); break;
		case OP_CPY_A:   op_cpX (cpu, cpu->y, AM_ABSOLUTE); break;
		case OP_CPY_Z:   op_cpX (cpu, cpu->y, AM_ZERO_PAGE); break;

		case OP_DEC_A:   op_dec (cpu, AM_ABSOLUTE); break;
		case OP_DEC_AX:  op_dec (cpu, AM_ABSOLUTE_X); break;
		case OP_DEC_Z:   op_dec (cpu, AM_ZERO_PAGE); break;
		case OP_DEC_ZX:  op_dec (cpu, AM_ZERO_PAGE_X); break;

		case OP_DEX_IMP: op_deX (cpu, REG_X); break;
		case OP_DEY_IMP: op_deX (cpu, REG_Y); break;

		case OP_EOR:     op_eor (cpu, AM_IMMEDIATE); break;
		case OP_EOR_A:   op_eor (cpu, AM_ABSOLUTE); break;
		case OP_EOR_AX:  op_eor (cpu, AM_ABSOLUTE_X); break;
		case OP_EOR_AY:  op_eor (cpu, AM_ABSOLUTE_Y); break;
		case OP_EOR_Z:   op_eor (cpu, AM_ZERO_PAGE); break;
		case OP_EOR_ZX:  op_eor (cpu, AM_ZERO_PAGE_X); break;
		case OP_EOR_IX:  op_eor (cpu, AM_INDIRECT_X); break;
		case OP_EOR_IY:  op_eor (cpu, AM_INDIRECT_Y); break;

		case OP_INC_A:   op_inc (cpu, AM_ABSOLUTE); break;
		case OP_INC_AX:  op_inc (cpu, AM_ABSOLUTE_X); break;
		case OP_INC_Z:   op_inc (cpu, AM_ZERO_PAGE); break;
		case OP_INC_ZX:  op_inc (cpu, AM_ZERO_PAGE_X); break;

		case OP_INX_IMP: op_inX (cpu, REG_X); break;
		case OP_INY_IMP: op_inX (cpu, REG_Y); break;

		case OP_JMP_A:   op_jmp (cpu, AM_ABSOLUTE); break;
		case OP_JMP_I:   op_jmp (cpu, AM_INDIRECT); break;

		case OP_JSR_A:   op_jsr (cpu); break;

		case OP_LDA:     op_ldX (cpu, REG_A, AM_IMMEDIATE); break;
		case OP_LDA_A:   op_ldX (cpu, REG_A, AM_ABSOLUTE); break;
		case OP_LDA_AX:  op_ldX (cpu, REG_A, AM_ABSOLUTE_X); break;
		case OP_LDA_AY:  op_ldX (cpu, REG_A, AM_ABSOLUTE_Y); break;
		case OP_LDA_Z:   op_ldX (cpu, REG_A, AM_ZERO_PAGE); break;
		case OP_LDA_ZX:  op_ldX (cpu, REG_A, AM_ZERO_PAGE_X); break;
		case OP_LDA_IX:  op_ldX (cpu, REG_A, AM_INDIRECT_X); break;
		case OP_LDA_IY:  op_ldX (cpu, REG_A, AM_INDIRECT_Y); break;

		case OP_LDX:     op_ldX (cpu, REG_X, AM_IMMEDIATE); break;
		case OP_LDX_A:   op_ldX (cpu, REG_X, AM_ABSOLUTE); break;
		case OP_LDX_AY:  op_ldX (cpu, REG_X, AM_ABSOLUTE_Y); break;
		case OP_LDX_Z:   op_ldX (cpu, REG_X, AM_ZERO_PAGE); break;
		case OP_LDX_ZY:  op_ldX (cpu, REG_X, AM_ZERO_PAGE_Y); break;

		case OP_LDY:     op_ldX (cpu, REG_Y, AM_IMMEDIATE); break;
		case OP_LDY_A:   op_ldX (cpu, REG_Y, AM_ABSOLUTE); break;
		case OP_LDY_AX:  op_ldX (cpu, REG_Y, AM_ABSOLUTE_X); break;
		case OP_LDY_Z:   op_ldX (cpu, REG_Y, AM_ZERO_PAGE); break;
		case OP_LDY_ZX:  op_ldX (cpu, REG_Y, AM_ZERO_PAGE_X); break;

		case OP_LSR_ACC: op_lsr (cpu, AM_ACCUMULATOR); break;
		case OP_LSR_A:   op_lsr (cpu, AM_ABSOLUTE); break;
		case OP_LSR_AX:  op_lsr (cpu, AM_ABSOLUTE_X); break;
		case OP_LSR_Z:   op_lsr (cpu, AM_ZERO_PAGE); break;
		case OP_LSR_ZX:  op_lsr (cpu, AM_ZERO_PAGE_X); break;

		case OP_NOP_IMP: op_nop (); break;

		case OP_ORA:     op_ora (cpu, AM_IMMEDIATE); break;
		case OP_ORA_A:   op_ora (cpu, AM_ABSOLUTE); break;
		case OP_ORA_AX:  op_ora (cpu, AM_ABSOLUTE_X); break;
		case OP_ORA_AY:  op_ora (cpu, AM_ABSOLUTE_Y); break;
		case OP_ORA_Z:   op_ora (cpu, AM_ZERO_PAGE); break;
		case OP_ORA_ZX:  op_ora (cpu, AM_ZERO_PAGE_X); break;
		case OP_ORA_IX:  op_ora (cpu, AM_INDIRECT_X); break;
		case OP_ORA_IY:  op_ora (cpu, AM_INDIRECT_Y); break;

		case OP_PHA_IMP: op_pha (cpu); break;
		case OP_PHP_IMP: op_php (cpu); break;
		case OP_PLA_IMP: op_pla (cpu); break;
		case OP_PLP_IMP: op_plp (cpu); break;

		case OP_ROL_ACC: op_rol (cpu, AM_ACCUMULATOR); break;
		case OP_ROL_A:   op_rol (cpu, AM_ABSOLUTE); break;
		case OP_ROL_AX:  op_rol (cpu, AM_ABSOLUTE_X); break;
		case OP_ROL_Z:   op_rol (cpu, AM_ZERO_PAGE); break;
		case OP_ROL_ZX:  op_rol (cpu, AM_ZERO_PAGE_X); break;

		case OP_ROR_ACC: op_ror (cpu, AM_ACCUMULATOR); break;
		case OP_ROR_A:   op_ror (cpu, AM_ABSOLUTE); break;
		case OP_ROR_AX:  op_ror (cpu, AM_ABSOLUTE_X); break;
		case OP_ROR_Z:   op_ror (cpu, AM_ZERO_PAGE); break;
		case OP_ROR_ZX:  op_ror (cpu, AM_ZERO_PAGE_X); break;

		case OP_RTI_IMP: op_rti (cpu); break;

		case OP_RTS_IMP:
			if (!op_rts (cpu)) {

				/* End of program */

				return FALSE;
			}
			break;

		case OP_SBC:     op_sbc (cpu, AM_IMMEDIATE); break;
		case OP_SBC_A:   op_sbc (cpu, AM_ABSOLUTE); break;
		case OP_SBC_AX:  op_sbc (cpu, AM_ABSOLUTE_X); break;
		case OP_SBC_AY:  op_sbc (cpu, AM_ABSOLUTE_Y); break;
		case OP_SBC_Z:   op_sbc (cpu, AM_ZERO_PAGE); break;
		case OP_SBC_ZX:  op_sbc (cpu, AM_ZERO_PAGE_X); break;
		case OP_SBC_IX:  op_sbc (cpu, AM_INDIRECT_X); break;
		case OP_SBC_IY:  op_sbc (cpu, AM_INDIRECT_Y); break;

		case OP_SEC_IMP: FLAG_SET (CPU_FLAG_C, TRUE); break;
		case OP_SED_IMP: FLAG_SET (CPU_FLAG_D, TRUE); break;
		case OP_SEI_IMP: FLAG_SET (CPU_FLAG_I, TRUE); break;

		case OP_STA_A:   op_stX (cpu, REG_A, AM_ABSOLUTE); break;
		case OP_STA_AX:  op_stX (cpu, REG_A, AM_ABSOLUTE_X); break;
		case OP_STA_AY:  op_stX (cpu, REG_A, AM_ABSOLUTE_Y); break;
		case OP_STA_Z:   op_stX (cpu, REG_A, AM_ZERO_PAGE); break;
		case OP_STA_ZX:  op_stX (cpu, REG_A, AM_ZERO_PAGE_X); break;
		case OP_STA_IX:  op_stX (cpu, REG_A, AM_INDIRECT_X); break;
		case OP_STA_IY:  op_stX (cpu, REG_A, AM_INDIRECT_Y); break;

		case OP_STX_A:   op_stX (cpu, REG_X, AM_ABSOLUTE); break;
		case OP_STX_Z:   op_stX (cpu, REG_X, AM_ZERO_PAGE); break;
		case OP_STX_ZY:  op_stX (cpu, REG_X, AM_ZERO_PAGE_Y); break;

		case OP_STY_A:   op_stX (cpu, REG_Y, AM_ABSOLUTE); break;
		case OP_STY_Z:   op_stX (cpu, REG_Y, AM_ZERO_PAGE); break;
		case OP_STY_ZX:  op_stX (cpu, REG_Y, AM_ZERO_PAGE_X); break;

		case OP_TAX_IMP: op_taX (cpu, REG_X); break;
		case OP_TAY_IMP: op_taX (cpu, REG_Y); break;
		case OP_TXA_IMP: op_tXa (cpu, REG_X); break;
		case OP_TYA_IMP: op_tXa (cpu, REG_Y); break;
		case OP_TSX_IMP: op_tsx (cpu); break;
		case OP_TXS_IMP: op_txs (cpu); break;

		default:
			op_not_implemented (cpu, op);
			return FALSE;
	}

	return TRUE;
}


