/* machine.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2017, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <assert.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"
#include "xlib/xstr.h"
#include "xlib/xfile.h"

#include "memory.h"
#include "ram.h"
#include "rom.h"
#include "video.h"
#include "keyboard.h"
#include "io.h"
#include "acia.h"
#include "profile.h"
#include "machine.h"
#include "bus.h"
#include "cpu.h"
#include "support.h"


#define TAPE_SIZE_MAX   (1024 * 64)


struct _Machine {

	const char  *progname;
	Profile     *profile;
	CPU         *cpu;
	Memory      *bus[ADDRESS_SPACE];
	RAM         *ram;
	xlist_t     *rom_list;
	Video       *video;
	Keyboard    *keyboard;
	ACIA        *acia;
	IO          *getc;
	IO          *putc;
	IO          *random;

	char        *fname_tape;
	char        *tape_data;

	xbool_t     quiet;

};


static void machine_tape_init (Machine *m, const char *fname);


static xbool_t
ram_load (Machine *m, const char *fname, int offset)
{
	char **split, *fname_tmp, *offset_tmp, *fullpath;
	int n;

	if (offset == -1) {
		split = xstr_split (fname, ",: ");
		for (n = 0; split[n] != NULL; n++);
		if (n != 2) {
			fprintf (stderr, "%s: Trouble parsing file info: '%s'\n",
			         m->progname, fname);
			xstr_freev (split);
			return FALSE;
		}
		fname_tmp = xstr_strip (xstr_dup (split[0]));
		offset_tmp = xstr_strip (xstr_dup (split[1]));
		offset = parse_ascii_integer (offset_tmp);
		xstr_freev (split);
		xfree (offset_tmp);
	}
	else {
		fname_tmp = xstr_dup (fname);
	}

	fullpath = xstr_join (DIR_SEPARATOR_S, m->profile->datadir,
	                      m->profile->fname, "profile", NULL);

	n = load_hex (m->progname, fullpath, (Memory*)m->ram, offset);

	xfree (fullpath);
	xfree (fname_tmp);

	return n > 0;
}


Machine*
machine_new (const char *progname, const char *fname_profile, int ram_size,
             const char *fname_load, int offset_load, const char *fname_tape,
             xbool_t quiet, xbool_t fullscreen)
{
	Machine *m;
	xlist_t *this;
	ROM *rom;

	m = xnew0 (Machine, 1);
	m->progname = progname;
	m->quiet = quiet;

#if DEBUG
	mem_set_quiet (m->quiet);
#endif

	bus_init (m->bus);

	/* Load profile */

	if ((m->profile = profile_new (progname, fname_profile)) == NULL) {
		machine_free (m);
		return NULL;
	}

	/* Add RAM */

	if (m->profile->ram_base >= 0) {
		m->ram = ram_new (m->profile, ram_size);
		bus_attach_device (m->bus, (Memory*)m->ram);
	}

	if (m->profile->load_list) {

		for (this = m->profile->load_list; this != NULL; this = this->next) {
			if (!ram_load (m, this->data, -1)) {
				machine_free (m);
				return NULL;
			}
		}
	}

	if (fname_load && !ram_load (m, fname_load, offset_load)) {
		machine_free (m);
		return NULL;
	}


	/* Load and add ROM(s) */

	for (this = m->profile->rom_list; this != NULL; this = this->next) {
		if ((rom = rom_new (m->profile, this->data)) != NULL) {
			bus_attach_device (m->bus, (Memory*)rom);
			xlist_append (&m->rom_list, rom);
		}
		else {
			machine_free (m);
			return NULL;
		}
	}

	/* Add video */

	if (m->profile->environment != ENVIRONMENT_STDIO
	    && m->profile->video_base >= 0) {

		if ((m->video = video_new (m->profile, fullscreen)) != NULL) {
			bus_attach_device (m->bus, (Memory*)m->video);
		}
		else {
			machine_free (m);
			return NULL;
		}

	}

	/* Add keyboard */

	if (m->profile->keyboard_type != KEYBOARD_TYPE_STDIO
	    && m->profile->keyboard_base >= 0) {

		m->keyboard = keyboard_new (m->profile);
		bus_attach_device (m->bus, (Memory*)m->keyboard);
	}

	/* Various I/O ports */

	if (m->profile->getc >= 0) {
		if ((m->getc = io_new (m->profile, m->profile->getc, 1)) != NULL) {
			bus_attach_device (m->bus, (Memory*)m->getc);
		}
	}
	if (m->profile->putc >= 0) {
		if ((m->putc = io_new (m->profile, m->profile->putc, 1)) != NULL) {
			bus_attach_device (m->bus, (Memory*)m->putc);
		}
	}
	if (m->profile->random >= 0) {
		if ((m->random = io_new (m->profile, m->profile->random, 1)) != NULL) {
			bus_attach_device (m->bus, (Memory*)m->random);
		}
	}

	/* ACIA */

	/* Set up tape player */
	if (m->profile->acia >= 0) {
		if (fname_tape) {
			machine_tape_init (m, fname_tape);
		}
		else {
			/* A dummy to avoid segfault on LOAD/SAVE without a tape */
			/* XXX: Reducing this to 1 byte might be fine (check first) */
			m->tape_data = xnew0 (char, TAPE_SIZE_MAX);
		}
	}

	if (m->profile->acia >= 0) {
		if ((m->acia = acia_new (m, m->profile->acia, m->tape_data)) != NULL) {
			bus_attach_device (m->bus, (Memory*)m->acia);
		}
	}

	/* Add CPU */

	m->cpu = cpu_new (m, m->progname, m->bus, m->profile->cpu_type,
	                  m->profile->vector_reset, m->profile->vector_nmi,
	                  m->profile->vector_irq);

	return m;
}


void
machine_free (Machine *m)
{
	if (m == NULL) {
		return;
	}

	ram_free (m->ram);

	if (m->rom_list) {
		xlist_t *this;
		for (this = m->rom_list; this != NULL; this = this->next) {
			rom_free (this->data);
		}
		xlist_free (&m->rom_list);
	}

	video_free (m->video);
	keyboard_free (m->keyboard);
	acia_free (m->acia);
	io_free (m->getc);
	io_free (m->putc);
	io_free (m->random);
	cpu_free (m->cpu);

	xfree (m->fname_tape);
	xfree (m->tape_data);

	profile_free (m->profile);

	xfree (m);
}


Profile*
machine_get_profile (Machine *m)
{
	return m->profile;
}


Keyboard*
machine_get_keyboard (Machine *m)
{
	return m->keyboard;
}


Video*
machine_get_video (Machine *m)
{
	return m->video;
}


void
machine_reset (Machine *m)
{
	keyboard_set_reset (m->keyboard, FALSE);
	cpu_signal_reset (m->cpu);
}


void
machine_toggle_stepping (Machine *m)
{
	keyboard_set_stepping (m->keyboard, FALSE);
	cpu_signal_toggle_stepping (m->cpu);
}


void
machine_step_now (Machine *m)
{
	keyboard_set_step_now (m->keyboard, FALSE);
	cpu_signal_step_now (m->cpu);
}


static int
machine_get_entry_point (Machine *m, int entry_point)
{
#ifdef DEBUG
	if (!m->quiet) {
		fprintf (stderr, "%s: entry point: ", m->progname);
	}
#endif

	if (entry_point == -1) {

		if (m->profile->startup != -1) {
#ifdef DEBUG
			if (!m->quiet) {
				fprintf (stderr, "profile->startup\n");
			}
#endif
			return m->profile->startup;
		}
		else if (m->profile->vector_reset != -1) {
#ifdef DEBUG
			if (!m->quiet) {
				fprintf (stderr, "profile->vector_reset\n");
			}
#endif
			return bus_read_word (m->bus, m->profile->vector_reset);
		}

	}

#ifdef DEBUG
	if (!m->quiet) {
		fprintf (stderr, "0x0000\n");
	}
#endif

	return 0x0000;
}


xbool_t
machine_start (Machine *m, int entry_point)
{
	xbool_t ret;

	entry_point = machine_get_entry_point (m, entry_point);

#if DEBUG
	if (!m->quiet) {
		fprintf (stderr, "%s: machine_start: %s\n", m->progname,
		         m->profile->simulation_name);
		fprintf (stderr, "%s: machine_start: reset_vector = ", m->progname);
		if (m->profile->vector_reset < 0) {
			fprintf (stderr, "-1\n");
		}
		else {
			fprintf (stderr, "%04X\n", m->profile->vector_reset);
		}
		fprintf (stderr, "%s: machine_start: entry_point  = %04X\n",
		         m->progname, entry_point);
	}
#endif

	ret = cpu_run (m->cpu, entry_point);

	if (!m->quiet) {
		fprintf (stderr, "\n");
		cpu_status_dump (stderr, m->cpu);
	}

	return ret;
}


static void
machine_tape_init (Machine *m, const char *fname)
{
	FILE *f;

	m->fname_tape = xstr_dup (fname);

	if ((f = fopen (m->fname_tape, "r")) == NULL) {
		fprintf (stderr, "%s: Failed to open '%s' (%s)\n",
		    m->progname, m->fname_tape, strerror (errno));
	}
	else {

		int len, c;

		m->tape_data = xnew0 (char, TAPE_SIZE_MAX);

		len = 0;

		while ((c = fgetc (f)) != EOF && len < TAPE_SIZE_MAX) {
			m->tape_data[len] = c;
			/* fprintf (stderr, "%02X ", (Byte)c); */
			len++;
		}
		/* fprintf (stderr, "\n"); */

		fclose (f);

		if (len >= TAPE_SIZE_MAX) {
			fprintf (stderr,
			    "%s: Tape is larger than the maximum %d bytes (%s)\n",
			    m->progname, TAPE_SIZE_MAX, m->fname_tape);
		}

	}

}


