/* profile.c - 6502 microcomputer emulator
 *
 * Copyright (C) 2017, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <assert.h>

#include "xlib/xutil.h"
#include "xlib/xmem.h"
#include "xlib/xstr.h"
#include "xlib/xlist.h"
#include "xlib/xfile.h"

#include "config.h"
#include "profile.h"
#include "support.h"


#define DEFAULT_SIM_NAME           PACKAGE
#define DEFAULT_CPU_TYPE           "6502"
#define DEFAULT_RAM_BASE           0x0000
#define DEFAULT_RAM_SIZE           4096
#define DEFAULT_RAM_CLEAR          FALSE
#define DEFAULT_SCREEN_WIDTH       -1
#define DEFAULT_SCREEN_HEIGHT      -1
#define DEFAULT_ENVIRONMENT        "stdio"
#define DEFAULT_VIDEO_TYPE         NULL
#define DEFAULT_VIDEO_BASE         -1
#define DEFAULT_VIDEO_SIZE         -1
#define DEFAULT_VIDEO_SCALE_W      2
#define DEFAULT_VIDEO_SCALE_H      2
#define DEFAULT_VIDEO_CHARSET      NULL
#define DEFAULT_VIDEO_CHAR_W       -1
#define DEFAULT_VIDEO_CHAR_H       -1
#define DEFAULT_VIDEO_CHAR_N       -1
#define DEFAULT_VIDEO_FG           "0xff, 0xCC, 0xCC"
#define DEFAULT_VIDEO_BG           "0x11, 0x11, 0x11"
#define DEFAULT_RESET_TIMEOUT      0
#define DEFAULT_KEYBOARD_TYPE      "stdio"
#define DEFAULT_KEYBOARD_BASE      -1
#define DEFAULT_ACIA               -1
#define DEFAULT_GETC               -1
#define DEFAULT_PUTC               -1
#define DEFAULT_RANDOM             -1
#define DEFAULT_VECTOR_NMI         -1
#define DEFAULT_VECTOR_RESET       -1
#define DEFAULT_VECTOR_IRQ         -1
#define DEFAULT_STARTUP            -1


typedef struct {

	char *label;
	char *value;

} ProfileVar;


static xbool_t  profile_load (Profile *p);
static void     free_var_list (xlist_t *list);
static char*    profile_get_str (const char *label, const char *fallback);
static int      profile_get_int (const char *label, int fallback);
static xbool_t  profile_get_bool (const char *label, xbool_t fallback);
static xlist_t* profile_get_str_list (const char *label, xlist_t *fallback);


static xlist_t *profile_vars = NULL;


Environment environment;


static void
profile_error (Profile *p, char *message)
{
	fprintf (stderr, "%s: %s\n", p->progname, message);
	xfree (message);

	return;
}


static void
profile_value_error (Profile *p, const char *message)
{
	profile_error (p, xstr_dup_printf ("%s in '%s'", message, p->fname));
	free_var_list (profile_vars);
	profile_vars = NULL;
	profile_free (p);
}


Profile*
profile_new (const char *progname, const char *fname)
{
	Profile *p;
	char *s;

	p = xnew0 (Profile, 1);

	p->progname = progname;
	p->rom_list = NULL;

	/* Remove trailing '/' if present, then get dirname and basename. */
	s = strrchr (fname, DIR_SEPARATOR_C);
	if (s != NULL && *(s + 1) == '\0') {
		char *fname_tmp = xstr_ndup (fname, s - fname);
		p->datadir = xfile_dirname (fname_tmp);
		p->fname = xfile_basename (fname_tmp);
		xfree (fname_tmp);
	}
	else {
		p->datadir = xfile_dirname (fname);
		p->fname = xfile_basename (fname);
	}

	if (p->fname != NULL) {
		if (!profile_load (p)) {
			free_var_list (profile_vars);
			profile_vars = NULL;
			profile_free (p);
			return NULL;
		}
	}

	p->simulation_name = profile_get_str ("name", DEFAULT_SIM_NAME);

	s = profile_get_str ("cpu", DEFAULT_CPU_TYPE);
	if (!parse_cpu_type (s, &p->cpu_type)) {
		profile_value_error (p, "Unknown cpu_type");
		xfree (s);
		return NULL;
	}
	xfree (s);

	p->ram_base = profile_get_int ("ram_base", DEFAULT_RAM_BASE);
	p->ram_size = profile_get_int ("ram_size", DEFAULT_RAM_SIZE);
	p->ram_clear = profile_get_bool ("ram_clear", DEFAULT_RAM_CLEAR);

	s = profile_get_str ("environment", DEFAULT_ENVIRONMENT);
	if (!parse_environment (s, &p->environment)) {
		profile_value_error (p, "Unknown environment");
		xfree (s);
		return NULL;
	}
	xfree (s);

	environment = p->environment;

	p->video_base = profile_get_int ("video_base", DEFAULT_VIDEO_BASE);
	p->video_size = profile_get_int ("video_size", DEFAULT_VIDEO_SIZE);
	p->video_scale_w = profile_get_int ("video_scale_width", DEFAULT_VIDEO_SCALE_W);
	p->video_scale_h = profile_get_int ("video_scale_height", DEFAULT_VIDEO_SCALE_H);
	p->video_charset_fname = profile_get_str ("video_charset", DEFAULT_VIDEO_CHARSET);
	p->video_char_width = profile_get_int ("video_char_width", DEFAULT_VIDEO_CHAR_W);
	p->video_char_height = profile_get_int ("video_char_height", DEFAULT_VIDEO_CHAR_H);
	p->video_char_count = profile_get_int ("video_char_count", DEFAULT_VIDEO_CHAR_N);
	p->screen_width = profile_get_int ("screen_width", DEFAULT_SCREEN_WIDTH);
	p->screen_height = profile_get_int ("screen_height", DEFAULT_SCREEN_HEIGHT);
	p->reset_timeout = profile_get_int ("reset_timeout", DEFAULT_RESET_TIMEOUT);

	s = profile_get_str ("video_fg", DEFAULT_VIDEO_FG);
	if (!parse_video_rgb (s, &p->video_fg[0], &p->video_fg[1], &p->video_fg[2])) {
		profile_value_error (p, "Unable to parse RGB video foreground colour");
	}
	xfree (s);

	s = profile_get_str ("video_bg", DEFAULT_VIDEO_BG);
	if (!parse_video_rgb (s, &p->video_bg[0], &p->video_bg[1], &p->video_bg[2])) {
		profile_value_error (p, "Unable to parse RGB video background colour");
	}
	xfree (s);

	s = profile_get_str ("keyboard_type", DEFAULT_KEYBOARD_TYPE);
	if (!parse_keyboard_type (s, &p->keyboard_type)) {
		profile_value_error (p, "Unknown keyboard_type");
		xfree (s);
		return NULL;
	}
	xfree (s);

	p->keyboard_base = profile_get_int ("keyboard_base", DEFAULT_KEYBOARD_BASE);
	p->acia = profile_get_int ("acia", DEFAULT_ACIA);
	p->getc = profile_get_int ("getc", DEFAULT_GETC);
	p->putc = profile_get_int ("putc", DEFAULT_PUTC);
	p->random = profile_get_int ("random", DEFAULT_RANDOM);
	p->vector_nmi = profile_get_int ("vector_nmi", DEFAULT_VECTOR_NMI);
	p->vector_reset = profile_get_int ("vector_reset", DEFAULT_VECTOR_RESET);
	p->vector_irq = profile_get_int ("vector_irq", DEFAULT_VECTOR_IRQ);
	p->startup = profile_get_int ("startup", DEFAULT_STARTUP);
	p->rom_list = profile_get_str_list ("load_rom", NULL);
	p->load_list = profile_get_str_list ("load_ram", NULL);

	if (p->video_size == -1) {
		p->video_size = p->screen_width * p->screen_height;
	}

	/* XXX: loaded values haven't been sanity checked */

	free_var_list (profile_vars);
	profile_vars = NULL;

	return p;
}


void
profile_free (Profile *p)
{
	xlist_t *this;

	if (p == NULL) {
		return;
	}

	xfree (p->fname);
	xfree (p->datadir);
	xfree (p->simulation_name);
	xfree (p->video_charset_fname);

	if (p->rom_list) {
		this = p->rom_list;
		while (this) {
			xfree (this->data);
			this = this->next;
		}
		xlist_free (&p->rom_list);
	}

	if (p->load_list) {
		this = p->load_list;
		while (this) {
			xfree (this->data);
			this = this->next;
		}
		xlist_free (&p->load_list);
	}

	xfree (p);
}


static void
free_var_list (xlist_t *list)
{
	ProfileVar *v;
	xlist_t *this;

	if (list == NULL) {
		return;
	}

	for (this = list; this != NULL; this = this->next) {

		v = this->data;

		xfree (v->label);
		xfree (v->value);
		xfree (v);

	}

	xlist_free (&list);
}


static ProfileVar*
profile_lookup (xlist_t *list, const char *label)
{
	ProfileVar *v;
	xlist_t *this;

	for (this = list; this != NULL; this = this->next) {
		v = this->data;
		if (strcasecmp (v->label, label) == 0) {
			return v;
		}
	}

	return NULL;
}


static char*
profile_get_str (const char *label, const char *fallback)
{
	ProfileVar *v;

	v = profile_lookup (profile_vars, label);
	if (v == NULL) {
		return xstr_dup (fallback);
	}

	return xstr_dup (v->value);
}


static int
profile_get_int (const char *label, int fallback)
{
	ProfileVar *v;

	v = profile_lookup (profile_vars, label);
	if (v == NULL) {
		return fallback;
	}

	return parse_ascii_integer (v->value);
}


static xbool_t
profile_get_bool (const char *label, xbool_t fallback)
{
	ProfileVar *v;

	v = profile_lookup (profile_vars, label);
	if (v == NULL) {
		return fallback;
	}

	return parse_ascii_bool (v->value);
}


static xlist_t*
profile_get_str_list (const char *label, xlist_t *fallback)
{
	ProfileVar *v;
	xlist_t *this;
	xlist_t *ret = NULL;

	for (this = profile_vars; this != NULL; this = this->next) {
		v = this->data;
		if (strcasecmp (v->label, label) == 0) {
			xlist_append (&ret, xstr_dup (v->value));
		}
	}
	if (ret == NULL) {
		/* Note: this is not a newly allocated copy, but the original */
		return fallback;
	}

	return ret;
}


static void
profile_set (const char *label, const char *value)
{
	ProfileVar *v;

	if (value == NULL || *value == '\0') {
		return;
	}

	v = profile_lookup (profile_vars, label);

	if (v != NULL
	    && strcasecmp (label, "load_rom")     /* allow multiple ROM loads */
	    && strcasecmp (label, "load_ram")) {  /* allow multiple RAM loads */

		/* Redefinition - just replace existing value */

		xfree (v->value);
		v->value = xstr_dup (value);

	}
	else {

		/* New definition */

		v = xnew (ProfileVar, 1);
		v->label = xstr_dup (label);
		v->value = xstr_dup (value);

		xlist_append (&profile_vars, v);

	}

}


static xbool_t
profile_load (Profile *p)
{
	size_t bufsize;
	ssize_t result;
	char *buf, **split, *fname;
	FILE *f;
	int line_num = 1;
	xbool_t okay = TRUE;


	fname = xstr_dup_printf ("%s%c%s%cprofile",
	                         p->datadir, DIR_SEPARATOR_C,
	                         p->fname, DIR_SEPARATOR_C);

	if ((f = fopen (fname, "r")) == NULL) {
		profile_error (p, xstr_dup_printf ("Unable to load machine profile '%s' (%s)",
		                                   fname, strerror (errno)));
		xfree (fname);
		return FALSE;
	}

	buf = NULL;

	while ((result = xfile_getline (&buf, &bufsize, f)) > 0) {

		if (*buf == '#' || *buf == ';' || *buf == '\n' || *buf == '\0') {
			line_num++;
			continue;
		}

		split = xstr_split (buf, "=");

		if (!split[0] || !split[1]) {
			xstr_freev (split);
			profile_error (p, xstr_dup_printf ("Error, line %d, %s", line_num, p->fname));
			line_num++;
			okay = FALSE;
			continue;
		}

		split[0] = xstr_strip (split[0]);
		split[1] = xstr_strip (split[1]);

		if (!str_valid (split[0], LABEL_PROFILE)) {
			profile_error (p, xstr_dup_printf ("Illegal label, line %d, %s", line_num, p->fname));
			xstr_freev (split);
			line_num++;
			okay = FALSE;
			continue;
		}

		profile_set (split[0], split[1]);

		xstr_freev (split);

		line_num++;
	}

	xfree (buf);
	fclose (f);

	xfree (fname);

	return okay;
}

